import sys
import numpy as np
import math, random, statistics
import simtk.unit as unit
from cg_openmm.build.cg_build import *
from cg_openmm.simulation.tools import *
from cg_openmm.utilities.iotools import *
from foldamers.cg_model.cgmodel import *

def random_sign(number):
        """
        Returns the provided 'number' with a random sign.

        :param number: The number to add a random sign (positive or negative) to
        :type number: float

        :returns:
          - number (float) - The provided number with a random sign added

        """

        # 0 = negative, 1 = positive
        random_int = random.randint(0,1)
        if random_int == 0: number = -1.0 * number

        return(number)

def first_bead(positions):
        """
        Determine if the provided 'positions' contain any particles (are the coordinates non-zero).

        :param positions: Positions for all beads in the coarse-grained model.
        :type positions: np.array( float * unit ( shape = num_beads x 3 ) )
        
        :returns:
         - first_bead (Logical) - Variable stating if the positions are all non-zero.

        """

        first_bead = True
    
        if str(positions.shape) == '(2,)':
          return(first_bead)
        else:
          for value in positions._value:
            if any(i != 0. for i in value):
              first_bead = False

        return(first_bead)


def get_move(trial_coordinates,move_direction,distance,bond_length,finish_bond=False):
        """
        Used to build random structures.  Given a set of input coordinates, this function attempts to add a new particle.

        :param trial_coordinates: Positions for a particle
        :type trial_coordinates: np.array( float * unit.angstrom ( length = 3 ) )

        :param move_direction: Cartesian direction in which we will attempt a particle placement, where: x=0, y=1, z=2. 
        :type move_direction: int

        :param distance: Current distance between the trial coordinates for the particle this function is positioning and the particle that it is branched from (bonded to).
        :type distance: `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :param bond_length: The distance to step before placing a new particle.
        :type bond_length: `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :param finish_bond: Logical variable determining how we will update the coordinates for this particle, default = False.  If set to "True", the "move" length will be the difference between "distance" and "bond_length".
        :type finish_bond: 

        :returns:
         - trial_coordinates (np.array( float * unit.angstrom (length=3) )) - Updated positions for the particle.

        """

        if distance.__gt__(bond_length):
          print("ERROR: The particle distance is larger than the bond length.")
          exit()

        # Determine the 'max_step_size' as the square root of the difference
        # between 'bond_length' and 'distance'
#        print("Bond length is "+str(bond_length))
#        print("Distance is "+str(distance))
        max_step_size = bond_length.__pow__(2.0).__sub__(distance.__pow__(2.0)).sqrt()

        # Add a random sign to 'max_step_size', to randomize our particle placement.
        sign_index = random.randint(0,1)
        if sign_index == 0:
          max_step_size = max_step_size.__neg__()


        # If we are "finishing the bond", then the "step" size is
        # the length (in the direction 'move_direction') required
        # so that 'distance' == 'bond_length'
        if finish_bond:

          # Calculate the step size as 'step' = sqrt('difference')
          step = max_step_size._value

        # If we aren't "finishing the bond", then the "step" size
        # is a random float in the range 0.0
        if not finish_bond:

          step = random.uniform(0.0,max_step_size._value)

        # Add this 'step' to the existing coordinates
        # print("The trial coordinates are: "+str(trial_coordinates))
        # print("The step size is: "+str(step))
        trial_coordinates[move_direction] = trial_coordinates[move_direction].__add__(unit.Quantity(step,trial_coordinates.unit))
        # print("The trial coordinates are: "+str(trial_coordinates))

        return(trial_coordinates)

def attempt_lattice_move(parent_coordinates,bond_length,move_direction_list):
        """
        Given a set of cartesian coordinates this function positions a new particle a distance of 'bond_length' away in a random direction.

        :param parent_coordinates: Positions for a single particle, away from which we will place a new particle a distance of 'bond_length' away.
        :type parent_coordinates: np.array( float * unit.angstrom ( length = 3 ) )

        :param bond_length: Bond length for all beads that are bonded.
        :type bond_length: `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :param move_direction_list: A list of cartesian directions (denoted by integers) that tracks the directions in which a particle placement has been attempted.
        :type move_direction_list: List( int )

        :returns:
          - trial_coordinates ( np.array( float * unit.angstrom ( length = 3 ) ) ) - The coordinates for a new, trial particle.
          - move_direction_list ( List(int) ) - A list of cartesian directions (denoted by integers) that tracks the directions in which a particle placement has been attempted.

        """

        units = bond_length.unit
        dist = unit.Quantity(0.0,units)

        # Assign the parent coordinates as the initial coordinates for a trial particle
        trial_coordinates = np.zeros([3]) * units
        try:
          parent_coordinates.unit
        except:
          parent_coordinates = parent_coordinates * units

        move_direction = random.randint(1,3)
        move_sign = random_sign(move_direction)
        while move_sign in move_direction_list:
         move_direction = random.randint(1,3)
         move_sign = random_sign(move_direction)
        move_direction_list.append(move_sign)
        if move_sign < 0:
         move_sign = -bond_length
        else:
         move_sign = bond_length
        #print("The trial coordinates are: "+str(trial_coordinates))
        #print("The parent coordinates are: "+str(parent_coordinates))
        trial_coordinates[:] = parent_coordinates[:]
        trial_coordinates[move_direction-1] = parent_coordinates[move_direction-1].__add__(move_sign)
        dist = distance(parent_coordinates,trial_coordinates)
        if round(dist._value,4) < round(bond_length._value,4):

           print("Error: particles are being placed at a distance different from the bond length")
           print("Bond length is: "+str(bond_length))
           print("The particle distance is: "+str(dist))
           print(ref)
           print(trial_coordinates)
           exit()

        return(trial_coordinates,move_direction_list)

def attempt_move(parent_coordinates,bond_length):
        """
        Given a set of cartesian coordinates, assign a new particle
        a distance of 'bond_length' away in a random direction.

        :param parent_coordinates: Positions for a single particle, away from which we will place a new particle a distance of 'bond_length' away.
        :type parent_coordinates: np.array( float * unit.angstrom ( length = 3 ) )

        :param bond_length: Bond length for all beads that are bonded.
        :type bond_length: `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :returns:
          - trial_coordinates ( np.array( float * unit.angstrom ( length = 3 ) ) ) - The coordinates for a new, trial
particle.

        """

        units = bond_length.unit
        dist = unit.Quantity(0.0,units)

        move_direction_list = []

        # Assign the parent coordinates as the initial coordinates for a trial particle

        trial_coordinates = parent_coordinates.__deepcopy__(memo={})
        ref = parent_coordinates.__deepcopy__(memo={})

        for direction in range(3):
            move_direction = random.randint(0,2)
            while move_direction in move_direction_list:
              move_direction = random.randint(0,2)

            if float(round(bond_length._value**2.0,4)-round(dist._value**2.0,4)) < 0.0:

              print("The bond length is: "+str(round(bond_length._value**2.0,4)))
              print("The distance is: "+str(round(dist._value**2.0,4)))
              print("The parent coordinates are: "+str(ref))
              print("The trial coordinates are: "+str(trial_coordinates))
              print("Error: new particles are not being assigned correctly.")
              exit()

            if direction == 2:
              trial_coordinates = get_move(trial_coordinates,move_direction,dist,bond_length,finish_bond=True)

            else:
              trial_coordinates = get_move(trial_coordinates,move_direction,dist,bond_length)

            move_direction_list.append(move_direction)

#            print("Parent coordinates are: "+str(ref))
#            print("Trial coordinates are: "+str(trial_coordinates))
            dist = distance(ref,trial_coordinates)
#            print(direction)
#            print(dist)

        if round(dist._value,4) < round(bond_length._value,4):

           print("Error: particles are being placed at a distance different from the bond length")
           print("Bond length is: "+str(bond_length))
           print("The particle distance is: "+str(dist))
           print(ref)
           print(trial_coordinates)
           exit()

        return(trial_coordinates)

def distances(interaction_list,positions):
        """
        Calculate the distances between all non-bonded particles in a model, given a list of particle interactions and particle positions.

        :param interaction_list: A list of non-bonded particle interactions
        :type interaction_list: List( [ int, int ] )

        :param positions: Positions for the particles in a coarse grained model.
        :type positions: np.array( float * unit.angstrom ( num_particles x 3 ) )

        :returns:
           - distance_list ( List( `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ) ) - A list of distances for the non-bonded interactions in the coarse grained model.

        """

        distance_list = []

#        print("In 'distances()' the positions are:"+str(positions))

        for interaction in interaction_list:
            if interaction[0] < len(positions) and interaction[1] < len(positions):
#             print("The distance between particles: "+str(interaction[0])+" and "+str(interaction[1])+" is "+str(distance(positions[interaction[0]],positions[interaction[1]])))
             distance_list.append(distance(positions[interaction[0]],positions[interaction[1]]))

        return(distance_list)

def collisions(positions,distance_list,distance_cutoff):
        """
        Determine if there are any collisions between non-bonded
        particles, where a "collision" is defined as a distance shorter than 'distance_cutoff'.

        :param positions: Positions for the particles in a coarse grained model.
        :type positions: np.array( float * unit.angstrom ( num_particles x 3 ) )

        :param distance_list: A list of distances.
        :type distance_list: List( `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ )

        :param distance_cutoff: The distance below which particles will be considered to have "collisions".
        :type distance_cutoff: `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :returns:
          - collision (Logical) - A variable indicating whether or not the model contains particle collisions.

        """

        collision = False

        #print("The nonbonded cutoff distance is: "+str(distance_cutoff._value))
        if len(distance_list) > 0:
          #print(distance_list)
          for distance in distance_list:

            if round(distance._value,4) < round(distance_cutoff._value,4):

              collision = True

        if positions.shape != '(3,)':
         for position_1_index in range(len(positions)):
          for position_2_index in range(len(positions)):
           if position_1_index != position_2_index:
            if all([positions[position_1_index][i] == positions[position_2_index][i] for i in range(len(positions[position_1_index]))]):
             #print("Detected a collision between particles "+str(position_1_index)+" and "+str(position_2_index))
             collision = True

        return(collision)

def assign_position_lattice_style(cgmodel,positions,distance_cutoff,parent_bead_index,bead_index):
        """
        Assign random position for a particle

        :param cgmodel: CGModel() class object.
        :type cgmodel: class

        :param positions: Positions for the particles in a coarse grained model.
        :type positions: np.array( float * unit.angstrom ( num_particles x 3 ) )

        :param distance_cutoff: The distance below which particles will be considered to have "collisions".
        :type distance_cutoff: `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :param parent_bead_index: The index of the particle from which we will bond a new particle, when assigning positions.
        :type parent_bead_index: int

        :param bead_index: The index of the particle for which the function will assign positions.
        :type bead_index: int

        :returns:
          - test_positions ( np.array( float * unit.angstrom ( num_particles x 3 ) ) ) - A set of positions for the updated model, including the particle that was just added.
          - success ( Logical ) - Indicates whether or not a particle was placed successfully.

        """
        #print("The first line positions are: "+str(positions))
        #print("The index for the bead we are trying to place is: "+str(bead_index))
        saved_positions = positions

        bond_list = cgmodel.get_bond_list()
        nonbonded_list = cgmodel.nonbonded_interaction_list

        #if str(positions.shape) == '(3,)':
          #parent_coordinates = positions
        #else:
          #print(parent_bead_index)
          #print(positions)
          #parent_coordinates = positions[parent_bead_index].__deepcopy__(memo={})

        success = False
        move_direction_list = []
        while len(move_direction_list) < 6 and not success:
           #print(len(move_direction_list))
           bond_length = cgmodel.get_bond_length(parent_bead_index,bead_index)
           #print("Calling attempt lattice move.")
           #print(move_direction_list)
           if str(positions.shape) == '(3,)':
            new_coordinates,move_direction_list = attempt_lattice_move(positions,bond_length,move_direction_list)
           else:
            #print(positions)
            #print(len(positions))
            #print(parent_bead_index)
            new_coordinates,move_direction_list = attempt_lattice_move(positions[parent_bead_index],bond_length,move_direction_list)
           if str(positions.shape) == '(3,)':
             test_positions = np.zeros([2,3]) * bond_length.unit
             test_positions[0] = positions
           else:
             test_positions = np.zeros([len(positions)+1,3]) * bond_length.unit
             test_positions[0:len(positions)] = positions[0:bead_index+2].__deepcopy__(memo={})
           #print("The provided positions were: "+str(positions))
           #print("The test positions before adding the trial coordinates are: "+str(test_positions))
           #print("The new coordinates are: "+str(new_coordinates))
           #print("The index for the bead we are trying to place is: "+str(bead_index))
           test_positions[bead_index] = new_coordinates
           #print("The test positions after adding the trial coordinates are: "+str(test_positions))
           #if len(test_positions) > 4: exit()
           nonbonded_distance_list = distances(nonbonded_list,test_positions)
           bonded_distance_list = distances(bond_list,test_positions)
           #print(nonbonded_list)
           #print(distance_list)
           if len(nonbonded_distance_list) == 0:
             success = True
             break
           if len(nonbonded_distance_list) > 0: 
            if not collisions(test_positions,nonbonded_distance_list,distance_cutoff) and not collisions(test_positions,bonded_distance_list,cgmodel.bond_lengths['bb_bb_bond_length']):
             success = True
             break 
        #print("Exited while loop in assign position lattice style.")       
        if success:
         if str(saved_positions.shape) != '(3,)':
          if all([test_positions[atom][index] == saved_positions[atom][index] for index in range(3) for atom in range(len(saved_positions))]) and len(test_positions) == len(saved_positions):
            print("ERROR: the input coordinates are the same as the output coordinates in 'assign_position_lattice_style()'.")
            print("The positions at the beginning of this function call were:"+str(saved_positions))
            print("The positions at the end of this function call are:"+str(positions))
            exit()
          else:
            positions = test_positions
            #print("Returing this set of positions: "+str(positions))
        #print(test_positions)
        return(test_positions,success)

def assign_position(positions,bond_length,distance_cutoff,parent_index,bead_index):
        """
        Assign random position for a bead

        :param positions: Positions for the particles in a coarse grained model.
        :type positions: np.array( float * unit.angstrom ( num_particles x 3 ) )

        :param bond_length: The distance to step when placing new particles.
        :type bond_length: `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :param distance_cutoff: The distance below which particles will be considered to have "collisions".
        :type distance_cutoff: `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :param parent_bead_index: The index of the particle from which we will bond a new particle, when assigning pos
itions.
        :type parent_bead_index: int

        :param bead_index: The index of the particle for which the function will assign positions.
        :type bead_index: int

        :returns:
          - positions ( np.array( float * unit.angstrom ( num_particles x 3 ) ) ) - A set of positions for the updated model, including the particle that was just added.
          - success ( Logical ) - Indicates whether or not a particle was placed successfully.

        """
        if bead_index == 1:
           success = True
           return(positions,success)

        units = bond_length.unit       
        if parent_index == -1:
               parent_index = bead_index - 1

        parent_coordinates = positions[parent_index-1]

        new_coordinates = unit.Quantity(np.zeros(3), units)
        success = False
        attempts = 0

        while not success:

           new_coordinates = attempt_move(parent_coordinates,bond_length)

           distance_list = distances(new_coordinates,positions)

           if not collisions(new_coordinates,distance_list,sigma): success = True

           if attempts > 100:
#            print([distance._value for distance in distance_list])
            return(positions,success)

           attempts = attempts + 1

        positions[bead_index-1] = new_coordinates

        return(positions,success)


def get_structure_from_library( cgmodel, high_energy=False, low_energy=False ):
        """
        Given a coarse grained model class object, this function retrieves
        a set of positions for the model from the 'foldamers' ensemble library, in:
        'foldamers/ensembles/${backbone_length}_${sidechain_length}_${sidechain_positions}'
        If this coarse grained model does not have an ensemble library, an 
        error message will be returned and positions at random with 'random_positions()'.

        cgmodel: CGModel() class object.

        :param high_energy: If set to 'True', this function will generate an ensemble of high-energy structures, default = False
        :type high_energy: Logical

        :param low_energy: If set to 'True', this function will generate an ensemble of low-energy structures, default = False
        :type low_energy: Logical

        :returns:
           - positions ( np.array( float * unit.angstrom ( num_particles x 3 ) ) ) - A set of coarse grained model positions.

        """
        ensemble_size = 10
        # In its current form this subroutine does not save the positions for random configurations we generate from heteropolymers.  It only saves the positions for homopolymers.

        if len(cgmodel.monomer_types) > 1:

          print("The foldamers ensemble library does not currently store conformations for polymers composed of more than one unique monomer.\n")
          print("The 'random_positions()' subroutine will be called instead, with 'use_library'=False.")
          positions = random_positions(cgmodel,use_library=False)
          return(positions)

        else:

          monomer_type = cgmodel.monomer_types[0]
          ensembles_directory = str(str(__file__.split('foldamers/utilities/util.py')[0])+"ensembles")
          if not os.path.exists(ensembles_directory):
            os.mkdir(ensembles_directory)
          model_directory = str(str(ensembles_directory)+"/"+str(cgmodel.polymer_length)+"_"+str(monomer_type['backbone_length'])+"_"+str(monomer_type['sidechain_length'])+"_"+str(monomer_type['sidechain_positions']))
          if not os.path.exists(model_directory):
            os.mkdir(model_directory)

          # We determine a suitable name for the ensemble directory by combining the 'bb_bb_bond_length', 'bb_sc_bond_length', and 'sc_sc_bond_length' into a single string:
          ens_str = [monomer_type['bond_lengths']['bb_bb_bond_length']._value,monomer_type['bond_lengths']['bb_sc_bond_length']._value,monomer_type['bond_lengths']['sc_sc_bond_length']._value]
          if not high_energy and not low_energy:
            ensemble_directory = str(str(model_directory)+"/bonds_"+str(ens_str[0])+"_"+str(ens_str[1])+"_"+str(ens_str[2]))
          if high_energy:
            ensemble_directory = str(str(model_directory)+"/bonds_"+str(ens_str[0])+"_"+str(ens_str[1])+"_"+str(ens_str[2])+"_high_energy")
          if low_energy:
            ensemble_directory = str(str(model_directory)+"/bonds_"+str(ens_str[0])+"_"+str(ens_str[1])+"_"+str(ens_str[2])+"_low_energy")

          generate_ensemble = False
          if not os.path.exists(ensemble_directory):
            os.mkdir(ensemble_directory)
            generate_ensemble = True
#            positions = random_positions(cgmodel,use_library=False)
          pdb_list = []
          energy_list = {'energy': [],'file_index': []}
          for file in os.listdir(ensemble_directory):
            if file.endswith('.pdb'):
              pdb_list.append(str(str(ensemble_directory)+"/"+str(file)))
              file_obj = open(str(str(ensemble_directory)+"/"+file),"r")
              #print(file_obj.readlines())
              if str('The OpenMM potential energy for this structure is:') in str(file_obj.readlines()):
                try:
                  energy = float(file_obj.readlines()[0].split(": ")[1].split(' ')[0])
                except:
                  energy = 9.9e9
              else:
                energy = 9.9e9
              energy_list['energy'].append(float(energy))
              file_index = int(file.split('cg')[1].split('.pdb')[0])
              energy_list['file_index'].append(file_index)
          if len(pdb_list) < ensemble_size:
            generate_ensemble = True

          if generate_ensemble:
#            print("The foldamers ensemble library only contains "+str(len(pdb_list))+" structures with these settings.\n")
#            print("The 'random_positions()' subroutine will be called instead, with 'use_library'=False,")
#            print("in order to generate a total of "+str(100)+" configurations for the database,")
#            print("before a specific configuration is chosen to assign random positions for this model.")
            if low_energy:
              total_iterations = 1000
            else:
              total_iterations = ensemble_size
            index = 1
            current_size = 1
            while index <= total_iterations and current_size < ensemble_size:
              file_name = str(ensemble_directory+"/cg"+str(current_size)+".pdb")
              if not os.path.exists(file_name):
                cgmodel.positions,cgmodel.simulation = random_positions(cgmodel,use_library=False)
                write_pdbfile_without_topology(cgmodel,file_name)
                cgmodel.topology = get_topology_from_pdbfile(file_name)
                cgmodel.system = build_system(cgmodel)
                simulation_time_step = 5.0 * unit.femtosecond
                positions_after,energy,simulation = minimize_structure(cgmodel.topology,cgmodel.system,cgmodel.positions,temperature=300.0 * unit.kelvin,simulation_time_step=5.0 * unit.femtosecond,total_simulation_time=0.1 * unit.picosecond,print_frequency=10)
                cgmodel.positions = positions_after
                write_pdbfile_without_topology(cgmodel,file_name)

                if low_energy and any(energy._value < energy_list['energy'][i] for i in range(len(energy_list['energy']))) and len(energy_list['energy']) <= ensemble_size:
                    highest_energy_index = energy_list['energy'].index(max(energy_list['energy']))
                    print("Replacing the structure in "+str(str(ensemble_directory+"/cg"+str(energy_list['file_index'][highest_energy_index])+".pdb")))
                    print("Which had an energy of: "+str(energy_list['energy'][highest_energy_index]))
                    print("with a new structure that has an energy of: "+str(energy._value)) 
                    energy_list['energy'][highest_energy_index] = energy._value
                    write_pdbfile_without_topology(cgmodel,str(ensemble_directory+"/cg"+str(energy_list['file_index'][highest_energy_index])+".pdb"),energy=energy)

                else:

                    if len(energy_list['energy']) <= ensemble_size:
                      for step in range(len(energy_list['energy'])):
                        file_name = str(ensemble_directory+"/cg"+str(step+1)+".pdb")
                        if not os.path.exists(file_name):
                          energy_list['energy'].append(float(energy._value))
                          energy_list['file_index'].append(step+1)
                      write_pdbfile_without_topology(cgmodel,file_name,energy=energy)
                if len(energy_list['energy']) > 1:
                    print("The average energy for the new ensemble is: "+str(statistics.mean(energy_list['energy'])))

                index = index + 1
                current_size = current_size + 1

              else:
                
                index = index + 1
                current_size = current_size + 1

        pdb_list = []
        for file in os.listdir(ensemble_directory):
          if file.endswith('.pdb'):
            pdb_list.append(str(str(ensemble_directory)+"/"+str(file)))
        if len(pdb_list) > 0:
          random_file = pdb_list[random.randint(0,len(pdb_list)-1)]
        cgmodel.positions = get_positions_from_pdbfile(random_file)
        try:
         cgmodel.simulation = simulation
        except:
         cgmodel.system = build_system(cgmodel)
         positions,energy,simulation = minimize_structure(cgmodel.topology,cgmodel.system,cgmodel.positions,temperature=300.0 * unit.kelvin,simulation_time_step=5.0 * unit.femtosecond,total_simulation_time=0.1 * unit.picosecond,print_frequency=10)

        return(positions)

def random_positions( cgmodel,max_attempts=1000,use_library=False,high_energy=False,low_energy=False,generate_library=False):
        """
        Assign random positions for all beads in a coarse-grained polymer.

        cgmodel: CGModel() class object.

        :param max_attempts: The maximum number of attempts to generate random positions a coarse grained model with the current parameters, default = 1000
        :type max_attempts: int
       
        :param use_library: A logical variable determining if a new random structure will be generated, or if an ensemble will be read from the 'foldamers' database, default = False
        :param use_library: Logical

        :param high_energy: If set to 'True', this function will generate an ensemble of high-energy structures, default = False
        :type high_energy: Logical

        :param low_energy: If set to 'True', this function will generate an ensemble of low-energy structures, default = False
        :type low_energy: Logical

        :param generate_library: If set to 'True', this function will save the poses that are generated to the 'foldamers' ensemble database.
        :type generate_libary: False

        :returns:
           - positions ( np.array( float * unit.angstrom ( num_particles x 3 ) ) ) - A set of coarse grained model positions.

        """

        if high_energy and low_energy:
            print("ERROR: Both 'high_energy' and 'low_energy' ensembles were requested in 'get_ensemble()'.  Please set only one of these variables to 'True', and call the function again.")
            exit()


        total_attempts = 0
        if use_library:
#            print("Attempting to find a suitable random starting configuration in the foldamers structural database.\n")
            positions,simulation = get_structure_from_library(cgmodel,high_energy=high_energy,low_energy=low_energy)
            return(positions)

        units = cgmodel.bond_lengths['bb_bb_bond_length'].unit
        positions = np.array([[0.0,0.0,0.0] for bead in range(cgmodel.num_beads)]) * units
        bond_list = cgmodel.get_bond_list()
        #print(bond_list)
        total_attempts = 0
        distance_cutoff =  0.99 * cgmodel.bond_lengths['bb_bb_bond_length']
        nonbonded_list = cgmodel.nonbonded_interaction_list
        lattice_style = True
        stored_positions = positions[0].__deepcopy__(memo={})
        while total_attempts < max_attempts and len(stored_positions) != len(positions):
         #print("Beginning build attempt #"+str(total_attempts+1)+".",flush=True)
         if str(positions.shape) == '(3,)':
           stored_positions = positions
         else:
           stored_positions = positions[0].__deepcopy__(memo={})
         for bond_index in range(len(bond_list)):
           bond = bond_list[bond_index]

           if lattice_style:

            if bond[1] > bond[0]:
              positions,placement = assign_position_lattice_style(cgmodel,stored_positions,distance_cutoff,bond[0],bond[1])
            if bond[0] > bond[1]:
              positions,placement = assign_position_lattice_style(cgmodel,stored_positions,distance_cutoff,bond[1],bond[0])

           else:

             positions,placement = assign_position(positions,cgmodel.bond_length,distance_cutoff,bond[1],bond[0])
 
           if not placement:

            if len(positions) < 3 and cgmodel.num_beads > 6:
              print("ERROR: having difficulty assigning random coordinates for this model,")
              print("even with a small number of particles ( <= 4 ).")
              print("The current positions are: "+str(positions))
              print("Please check the parameter values and other settings for your coarse grained model.")
              print("Then check the 'random_positions()' subroutine for errors.")
              exit()
            if bond_index > 4: start = bond_index - 2
            else: start = 2
            #print("Restarting random configuration build from bond #"+str(start)) 
            #print("This attempt produced a model with length= "+str(len(positions)))
            total_attempts = total_attempts + 1
            stored_positions = positions[0:min(bond_list[start])].__deepcopy__(memo={})
            #print("The new positions are:"+str(stored_positions)+"\n")
            break
           if placement:
            if str(stored_positions.shape) != '(3,)':
             if int(len(stored_positions)) != int(len(positions)-1):
              print("ERROR: coordinates are not being added correctly in the 'random_positions()' subroutine.")
              print("The positions before appension are:"+str(stored_positions))
              print("The positions after appension are:"+str(positions))
              exit()
             #print(positions)
             #print(max(bond_list[bond_index])+1)
             stored_positions = positions[0:max(bond_list[bond_index])+1].__deepcopy__(memo={})
            else:
             #print(stored_positions)
             #print(positions)
             stored_positions = positions
#            bond_index = index + 1
          #if int(len(stored_positions)) == int(len(positions)): break
        distance_list = distances(nonbonded_list,positions)
        if len(distance_list) == 0 or not collisions(positions,distance_list,distance_cutoff):
          cgmodel.positions = positions
          cgmodel.topology = build_topology(cgmodel,use_pdbfile=True)
          cgmodel.system = build_system(cgmodel)
          #write_pdbfile_without_topology(cgmodel,"random_pose.pdb")
          positions,energy,simulation = minimize_structure(cgmodel.topology,cgmodel.system,positions,temperature=300.0 * unit.kelvin,simulation_time_step=5.0 * unit.femtosecond,total_simulation_time=0.1 * unit.picosecond,print_frequency=10)
        
          #print("No collisions detected.")
          return(positions)

        else:

          print("Error: A model was successfully built, however,")
          print("particle collisions were detected.\n")
          print("The shortest nonbonded particle distance is:")
          print(str(min(distance_list)))
          #collision = nonbonded_list[distance_list.index(min(distance_list))]
          print("The nonbonded particle cutoff distance used for")
          print("random structure generation is set to:"+str(distance_cutoff))
          print("the positions are: "+str(positions))
          simulation = None
#          print("This will continue until the user issues a disruption command with the keyboard. (Ctrl + c)")
        return(positions)

def distance(positions_1,positions_2):
        """
        Calculate the distance between two particles.

        :param positions_1: Positions for a particle
        :type positions_1: `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :param positions_2: Positions for a particle
        :type positions_2: `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :returns:
           - distance ( `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ) - The distance between the provided particles.

        """

        direction_comp = np.zeros(3) * positions_1.unit

        for direction in range(len(direction_comp)):
          direction_comp[direction] = positions_1[direction].__sub__(positions_2[direction])

        direction_comb = np.zeros(3) * positions_1.unit.__pow__(2.0)
        for direction in range(3):
          direction_comb[direction] = direction_comp[direction].__pow__(2.0)

        sqrt_arg = direction_comb[0].__add__(direction_comb[1]).__add__(direction_comb[2])

        value = math.sqrt(sqrt_arg._value)
        units = sqrt_arg.unit.sqrt()
        distance = unit.Quantity(value=value,unit=units)

        return(distance)


def distance_matrix(positions):
        """
        Construct a matrix of the distances between an input array of particles.

        :param positions: Positions for an array of particles.
        :type positions: np.array( float * unit.angstrom ( num_particles x 3 ) )

        :returns:
         - distance_matrix (np.array(num_particles x num_particles)) - Matrix containing the distances between all beads.
        """

        distance_matrix = np.array([[0.0 for index in range(0,len(positions))] for index in range(0,len(positions))])

        for index_1 in range(0,len(positions)):
          for index_2 in range(0,len(positions)):
            distance_matrix[index_1][index_2] = get_distance(positions[index_1],positions[index_2])

        return(distance_matrix)   



import numpy as np
import matplotlib.pyplot as pyplot
import mdtraj as md
import simtk.unit as unit
from statistics import mean
from foldamers.cg_model.cgmodel import basic_cgmodel
from cg_openmm.simulation.tools import *
#from genetic_selection import GeneticSelectionCV



def optimize_model_parameter(optimization_target):
        """
        """
        
        return

def get_fwhm_symmetry(C_v,T_list):
        """
        """

        symmetry_list = []
        fwhm_list = []

        for i in range(1,len(C_v)-1):
          if C_v[i] >= C_v[i-1] and C_v[i] >= C_v[i+1]:
            max_value = C_v[i]
            max_value_T = T_list[i]
            half_max = 0.5*max_value

            for j in range(i,0,-1):
              if C_v[j] < half_max:
                break
            # interpolate to get lower HM
            lower_C = C_v[j]
            lower_T = T_list[j]
            upper_C = C_v[j+1]
            upper_T = T_list[j+1]

            delta_T = upper_T - lower_T
            delta_C = upper_C - lower_C            
            lower_delta_C = half_max - lower_C
            
            delta_hm = lower_delta_C/delta_C

            lower_hm_T = delta_hm*delta_T

            for j in range(i,len(C_v)):
              if C_v[j] < half_max:
                break
            # interpolate to get lower HM
            lower_C = C_v[j]
            lower_T = T_list[j-1]
            upper_C = C_v[j-1]
            upper_T = T_list[j]

            delta_T = upper_T - lower_T
            delta_C = upper_C - lower_C
            upper_delta_C = C_v[j-1] - half_max

            delta_hm = upper_delta_C/delta_C

            upper_hm_T = delta_hm*delta_T       

            fwhm = upper_hm_T - lower_hm_T
            fwhm_list.append(fwhm)

            lower_delta = max_value_T - lower_hm_T
            upper_delta = upper_hm_T - max_value_T
           
            symmetry = abs(lower_delta-upper_delta)

            symmetry_list.append(symmetry)

        print("Symmetry = "+str(symmetry_list))
        print("FWHM = "+str(fwhm_list))

        symmetry = mean(symmetry_list)
        fwhm = mean(fwhm_list)

        return(symmetry,fwhm)


def get_num_maxima(C_v):
        """
        """

        maxima = 0

        for i in range(1,len(C_v)):
          if C_v[i] >= C_v[i-1] and C_v[i] >= C_v[i+1]:
            maxima = maxima + 1

        return(maxima)


def calculate_C_v_fitness(C_v,T_list):
        """
        """
        num_maxima = get_num_maxima(C_v)
        symmetry = get_fwhm_symmetry(C_v,T_list)

        num_maxima_weight = 0.33
        symmetry_wieght = 0.33
        fwhm_weight = 0.33

        maxima_fitness = math.exp( - ( num_maxima - 1 ) * num_maxima_weight )
        symmetry_fitness = math.exp( - ( abs(symmetry - 1) ) * symmetry_weight )          
        fwhm_fitness = math.exp( - ( fwhm ) * fwhm_weight )

        fitness = maxima_fitness * symmetry_fitness * fwhm_fitness

        return(fitness)


def optimize_heat_capacity(cgmodel):
        """
        """

        return

def optimize_lj(cgmodel,base_epsilon=0.0,sigma_attempts=3,epsilon_attempts=3):
        """
        Optimize the Lennard-Jones interaction potential parameters (sigma and epsilon, for all interaction types) in the model defined by a cgmodel() class object, using a combination of replica exchange simulations and re-weighting techniques.

        Parameters
        ----------

        :param cgmodel: CGModel() class object, default = None
        :type cgmodel: class.

        :param sigma_attempts: 

        Returns
        -------

        cgmodel: CGModel() class object.

        """

        # Set variable model settings
        base_sigma = cgmodel.sigmas['bb_bb_sigma'] # Lennard-Jones interaction distance
        base_epsilon = cgmodel.epsilons['bb_bb_epsilon'] # Lennard-Jones interaction strength
        sigma_list = [(base_sigma).__add__(i * base_sigma.unit) for i in [ j * 0.2 for j in range(-2,3,1)]]
        epsilon_list = [(base_epsilon).__add__(i * base_epsilon.unit) for i in [ j * 0.2 for j in range(-1,3,1)]]
        sigma_epsilon_list = np.zeros((len(sigma_list),len(epsilon_list)))

        for sigma_index in range(len(sigma_list)):
          for epsilon_index in range(len(epsilon_list)):
            sigma = sigma_list[sigma_index]
            epsilon = epsilon_list[epsilon_index]
            print("Evaluating the energy for a model with:")
            print("sigma="+str(sigma)+" and epsilon="+str(epsilon))
            # Build a coarse grained model
            cgmodel = basic_cgmodel(polymer_length=polymer_length, backbone_length=backbone_length, sidechain_length=sidechain_length, sidechain_positions=sidechain_positions, mass=mass, sigma=sigma, epsilon=epsilon, bond_length=bond_length)

            # Run replica exchange simulations with this coarse grained model.
            replica_energies,replica_temperatures = replica_exchange(cgmodel.topology,cgmodel.system,cgmodel.positions)
            print(replica_energies)

        

        return(cgmodel)

def optimize_parameter(cgmodel,optimization_parameter,optimization_range_min,optimization_range_max,steps=None):
        """
        """
        if steps == None: steps = 100
        step_size = ( optimization_range_max - optimization_range_min ) / 100
        parameter_values = [step*step_size for step in range(1,steps)]
        potential_energies = []
        for parameter in parameter_values:
          cgmodel.optimization_parameter = parameter
          positions,potential_energy,time_step =  minimize_structure(cgmodel.topology,cgmodel.system,cgmodel.positions,temperature=300.0 * unit.kelvin,simulation_time_step=None,total_simulation_time=1.0 * unit.picosecond,output_pdb='minimum.pdb',output_data='minimization.dat',print_frequency=1)
          potential_energies.append(potential_energy)

        best_value = min(potential_energies)

        return(best_value,potential_energies,parameter_values)



import os, subprocess
import numpy as np
import simtk.unit as unit
from statistics import mean
from scipy.stats import linregress
from scipy import spatial
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
from foldamers.utilities.util import *
from foldamers.utilities.iotools import write_pdbfile_without_topology

def get_native_contacts(cgmodel,native_structure,native_contact_distance_cutoff):
        """
        Given a coarse grained model, positions for that model, and positions for the native structure, this function calculates the fraction of native contacts for the model.

        :param cgmodel: CGModel() class object
        :type cgmodel: class

        :param native_structure: Positions for the particles in a coarse grained model.
        :type native_structure: np.array( float * unit.angstrom ( num_particles x 3 ) )

        :param native_contact_distance_cutoff: The maximum distance for two nonbonded particles that are defined as "native",default=None
        :type native_contact_distance_cutoff: `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :returns:
          - native_contact_list - A list of the nonbonded interactions whose inter-particle distances are less than the 'native_contact_cutoff_distance'.

        """

        nonbonded_interaction_list = cgmodel.nonbonded_interaction_list
        native_structure_distances = distances(nonbonded_interaction_list,native_structure)
        native_contact_list = []

        for interaction in range(len(nonbonded_interaction_list)):
          if native_structure_distances[interaction].__lt__(native_contact_distance_cutoff):
            native_contact_list.append(nonbonded_interaction_list[interaction])

        return(native_contact_list)

def get_number_native_contacts(cgmodel,native_structure,native_contact_distance_cutoff):
        """
        Given a coarse grained model, positions for that model, and positions for the native structure, this function calculates the fraction of native contacts for the model.

        :param cgmodel: CGModel() class object
        :type cgmodel: class

        :param native_structure: Positions for the particles in a coarse grained model.
        :type native_structure: np.array( float * unit.angstrom ( num_particles x 3 ) )

        :param native_contact_cutoff_distance: The maximum distance for two nonbonded particles that are defined as "native",default=None
        :type native_contact_cutoff_distance: `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :returns:
          - contacts - The number of nonbonded interactions that are considered 'native'.

        """

        native_contact_list = get_native_contacts(cgmodel,native_structure,native_contact_distance_cutoff)

        contacts = len(native_contact_list)

        return(contacts)

def fraction_native_contacts(cgmodel,positions,native_structure,native_structure_contact_distance_cutoff=None,native_contact_cutoff_ratio=None):
        """
        Given a coarse grained model, positions for that model, and positions for the native structure, this function calculates the fraction of native contacts for the model.

        :param cgmodel: CGModel() class object
        :type cgmodel: class

        :param positions: Positions for the particles in a coarse grained model.
        :type positions: np.array( float * unit.angstrom ( num_particles x 3 ) )

        :param native_structure: Positions for the native structure.
        :type native_structure: np.array( float * unit.angstrom ( num_particles x 3 ) )

        :param native_structure_contact_distance_cutoff: The distance below which two nonbonded, interacting particles that are defined as "native contact",default=None
        :type native_structure_contact_distance_cutoff: `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :param native_contact_cutoff_ratio: The distance below which two nonbonded, interacting particles in a non-native pose are assigned as a "native contact", as a ratio of the distance for that contact in the native structure, default=None
        :type native_contact_cutoff_ratio: `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :returns:
          - Q ( float ) - The fraction of native contacts for the comparison pose.

        """

        if native_structure_contact_distance_cutoff == None:
          native_structure_contact_distance_cutoff = 1.1*cgmodel.get_sigma(0)

        if native_contact_cutoff_ratio == None:
          native_contact_cutoff_ratio = 1.05

        native_contact_list = get_native_contacts(cgmodel,native_structure,native_structure_contact_distance_cutoff)

        total_native_interactions = get_number_native_contacts(cgmodel,native_structure,native_structure_contact_distance_cutoff)

        if total_native_interactions == 0:
          print("ERROR: there are 0 'native' interactions with the current cutoff distance.")
          print("Try increasing the 'native_structure_contact_distance_cutoff'")
          exit()

        native_contact_distances = distances(native_contact_list,native_structure)
        
        current_structure_distances = distances(native_contact_list,positions)

        current_structure_native_contact_list = []
        for interaction in range(len(native_contact_list)):
          if current_structure_distances[interaction].__lt__(native_contact_cutoff_ratio*native_contact_distances[interaction]):
            current_structure_native_contact_list.append(native_contact_list[interaction])
        current_structure_number_native_interactions = len(current_structure_native_contact_list)
        Q = current_structure_number_native_interactions / total_native_interactions
        return(Q)

def optimize_Q(cgmodel,native_structure,ensemble):
        """
        Given a coarse grained model and a native structure as input

        :param cgmodel: CGModel() class object
        :type cgmodel: class

        :param native_structure: Positions for the native structure.
        :type native_structure: np.array( float * unit.angstrom ( num_particles x 3 ) )

        :param ensemble: A list of poses that will be used to optimize the cutoff distance for defining native contacts
        :type ensemble: List(positions(np.array(float*simtk.unit (shape = num_beads x 3))))

        :returns:
          - native_structure_contact_distance_cutoff ( `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ) - The ideal distance below which two nonbonded, interacting particles should be defined as a "native contact"
        """

        cutoff_list = [(0.95+i*.01)*cgmodel.get_sigma(0) for i in range(30)]

        cutoff_Q_list = []
        for cutoff in cutoff_list:
          Q_list = []
          for pose in ensemble:
            Q = fraction_native_contacts(cgmodel,pose,native_structure,native_structure_contact_distance_cutoff=cutoff)
            Q_list.append(Q)

          mean_Q = mean(Q_list)
          cutoff_Q_list.append(mean_Q)

        cutoff_Q_list.index(max(cutoff_Q_list))

        native_structure_contact_distance_cutoff = cutoff_Q_list.index(max(cutoff_Q_list))

        return(native_structure_contact_distance_cutoff)


def get_helical_parameters(cgmodel):
        """
        Given a coarse grained model as input, this function uses the `kHelios software package <https://pubs.acs.org/doi/10.1021/acs.jcim.6b00721>`_ to analyze the helical properties of the model.

        :param cgmodel: CGModel() class object
        :type cgmodel: class

        :returns:
          - pitch ( float ) - The distance between monomers in adjacent turns of a helix
          - radius ( float ) - The radius of the helix
          - monomers_per_turn ( float ) - The number of monomrs per turn of the helix
          - residual ( float ) - The average distance of all backbone particles from a circle projected onto the x-y plane.  Used to determine the accuracy of the helical axis, as fit to the input data.  Units are in Angstroms.

        .. warning:: This function requires a pre-installed version of `kHelios <https://pubs.acs.org/doi/10.1021/acs.jcim.6b00721>`_ .  Because kHelios is formatted to accept input job scripts, this function writes and executes a job script for kHelios.  In order to function properly, the user must redefine the 'helios_path' variable for their system.

        """
        helios_path = str('/home/gmeek/Foldamers/foldamers/foldamers/parameters/helios.o')
        cgmodel = orient_along_z_axis(cgmodel)
        write_pdbfile_without_topology(cgmodel,"temp_pitch.pdb")
        kHelix_run_file = "run_kHelix.sh"
        file = open(kHelix_run_file,"w")
        file.write('#!/bin/bash\n')
        file.write('\n')
        file.write('cat > input << EOF\n')
        file.write('inputhelix $1\n')
        file.write('helixout_name kHelix.out\n')
        file.write('coord_type 1\n')
        file.write('num_grid 20\n')
        file.write('natoms '+str(round(cgmodel.num_beads/2))+'\n')
        file.write('nframes 1\n')
        file.write('grid_phi_beg 0\n')
        file.write('grid_phi_end 20\n')
        file.write('grid_theta_beg 0\n')
        file.write('grid_theta_end 20\n')
        file.write('helix_atom_names X1\n')
        file.write('print_to_plot 1\n')
        file.write('EOF\n')
        file.write(str(helios_path)+' input\n')
        #file.write('done\n')        
        file.close()
        subprocess.run(['chmod','+x','run_kHelix.sh'])
        subprocess.run([str(str(os.getcwd())+"/"+str(kHelix_run_file)),"temp_pitch.pdb",">","helios_output"])
        #os.remove("helios_output")
        file = open("kHelix.out",mode="r")
        output = file.readlines()
        line_index = 1
        for line in output:
          if line_index == 43:
            residual = line.split()[2]
            radius = line.split()[3]
            pitch = line.split()[4]
            sweep = float(line.split()[5])
            monomers_per_turn = cgmodel.polymer_length/(sweep/360.0)
            break
          line_index = line_index + 1
          
        return(pitch,radius,monomers_per_turn,residual)

def orient_along_z_axis(cgmodel,plot_projections=False):
        """
        Given a coarse grained model as input, this function orients the model along the z-axis.

        :param cgmodel: CGModel() class object
        :type cgmodel: class

        :param plot_projections: Variable indicating whether or not to plot intermediate projections/operations during identification of a helical axis.

        :returns:
           - cgmodel ( class ) - CGModel() class object, with positions oriented so that the helical axis is along the z-axis

        """
        positions = np.array([[float(i.in_units_of(unit.angstrom)._value) for i in position] for position in cgmodel.positions])
        # 1) Get the backbone particle positions
        backbone_positions = []
        for particle in range(len(cgmodel.positions)):
          if cgmodel.get_particle_type(particle) == "backbone":
            backbone_positions.append(cgmodel.positions[particle])
        backbone_positions = np.array([[float(i.in_units_of(unit.angstrom)._value) for i in coord] for coord in backbone_positions])
        # 2) Project the backbone positions onto the (x,y) plane
        xy_projected_positions = backbone_positions
        x_data = []
        y_data = []
        for position in xy_projected_positions:
          position[2] = 0.0
          x_data.append(position[0])
          y_data.append(position[1])
        # 3) Calculate the best fit line for these projected positions
        slope,intercept,r,p,std_err=linregress(np.array([x for x in x_data]),np.array([y for y in y_data]))

        if plot_projections:
        # Plot this projected data, as well as the best fit line   
          file_name = "xy_projection.png"
          figure = pyplot.figure(1)
          x_data = np.array([x for x in x_data])
          y_data = np.array([y for y in y_data])
          pyplot.xlabel("x")
          pyplot.ylabel("y")
          pyplot.scatter(x_data,y_data)
          x = np.linspace(min(x_data),max(x_data),100)
          pyplot.plot(x,slope*x+intercept,label=str("y="+str(round(slope,2))+"x+"+str(round(intercept,2))))
          pyplot.legend()
          pyplot.savefig(file_name)
          pyplot.show()
          pyplot.close()

        # 4) Rotate the coordinate system so that this line is oriented along the x-axis
        # Calculate angle from z-axis:
        z_axis_angle = np.arctan(slope)
        z_axis_rotation_matrix = spatial.transform.Rotation.from_euler('xyz',[0.,0.,z_axis_angle], degrees=False)
        x_oriented_positions = z_axis_rotation_matrix.apply(positions)

        # 5) Project the positions onto the (x,z) plane
        xz_projected_positions = backbone_positions
        x_data = []
        z_data = []
        for position_index in range(len(xz_projected_positions)):
          xz_projected_positions[position_index][1] = 0.0
          x_data.append(positions[position_index][0])
          z_data.append(positions[position_index][2])

        # 6) Calculate the best fit line for these projected positions
        slope,intercept,r,p,std_err=linregress(np.array([x for x in x_data]),np.array([z for z in z_data]))

        if plot_projections:
        # Plot this projected data, as well as the best fit line   
          file_name = "xz_projection.png"
          figure = pyplot.figure(1)
          x_data = np.array([x for x in x_data])
          z_data = np.array([z for z in z_data])
          pyplot.xlabel("x")
          pyplot.ylabel("z")
          pyplot.scatter(x_data,z_data)
          x = np.linspace(min(x_data),max(x_data),100)
          pyplot.plot(x,slope*x+intercept,label=str("z="+str(round(slope,2))+"x+"+str(round(intercept,2))))
          pyplot.legend()
          pyplot.savefig(file_name)
          pyplot.show()
          pyplot.close()


        # 7) Rotate the coordinate system so that this line is oriented along the x-axis
        # Calculate angle from y-axis:
        y_axis_angle = np.arctan(slope)
        y_axis_rotation_matrix = spatial.transform.Rotation.from_euler('xyz', [0.,y_axis_angle,0.], degrees=False)

        new_positions = y_axis_rotation_matrix.apply(x_oriented_positions)

        # 8) For comparison with kHelios output, rotate the molecule again so that the helical (x) axis is oriented along the z axis instead.

        z_axis_angle = 3.1415/2.0
        z_axis_rotation_matrix = spatial.transform.Rotation.from_euler('xyz',[0.,z_axis_angle,0.0], degrees=False)
        final_positions = z_axis_rotation_matrix.apply(new_positions)

        cgmodel.positions = unit.Quantity(final_positions,unit.angstrom)

        file = open("after_rotation.pdb","w")
        PDBFile.writeFile(cgmodel.topology,cgmodel.positions,file=file)

        return(cgmodel)

def show_helical_fit(cgmodel):
        """
        Given a coarse grained model containing positions, this function performs a helical fit for the backbone particles with `kHelios <https://pubs.acs.org/doi/10.1021/acs.jcim.6b00721>`_ , and uses 'matplotlib' to display attributes of the helical fit.
        """
        # 1) Get the backbone particle positions
        positions = np.array([[float(i.in_units_of(unit.angstrom)._value) for i in position] for position in cgmodel.positions])
        backbone_positions = []
        for particle in range(len(cgmodel.positions)):
          if cgmodel.get_particle_type(particle) == "backbone":
            backbone_positions.append(cgmodel.positions[particle])
        backbone_positions = np.array([[float(i.in_units_of(unit.angstrom)._value) for i in coord] for coord in backbone_positions])
        c = backbone_positions

        curves = [c]
        labels = ['helix (unrotated)']
        for i in range(len(curves)):
         fig = plt.figure(i)
         curve = curves[i]
         label = labels[i]
         ax = fig.gca(projection='3d')
         ax.plot(curve[:,0], curve[:,1], curve[:,2], label=label)
         ax.legend()
         plt.xlabel('x')
         plt.ylabel('y')
         #plt.zlabel('z') # not defined?
         plt.show()


        return

def calculate_p2(cgmodel):
        """
        Given a coarse grained model containing positions, this function returns the `'P2' <http://cmt.dur.ac.uk/sjc/thesis_dlc/node19.html>`_ orientational ordering parameter value for the current pose.

        .. warning:: By default, 'P2' is evaluated using the positions for only the backbone particles.

        :param cgmodel: CGModel() class object
        :type cgmodel: class

        :returns:
         - p2 ( float ) - The value for the 'P2' orientational ordering parameter.
        """
        positions = np.array([[float(i.in_units_of(unit.angstrom)._value) for i in position] for position in cgmodel.positions])
        # 1) Get the backbone particle positions
        backbone_positions = []
        for particle in range(len(cgmodel.positions)):
          if cgmodel.get_particle_type(particle) == "backbone":
            backbone_positions.append(cgmodel.positions[particle])
        backbone_positions = np.array([[float(i.in_units_of(unit.angstrom)._value) for i in coord] for coord in backbone_positions])
        
        c = backbone_positions

        u = np.diff(c,axis=0)
        for ui in u:
            ui /= np.sqrt(np.dot(ui,ui))

        Q = np.zeros([3,3])
        for ui in u:
            Q += 1.5*np.outer(ui,ui)
        Q /= len(u)
        Q -= 0.5*np.eye(3)
        vals, vecs = np.linalg.eig(Q)
        p2 = np.mean(np.dot(u,vecs),axis=0)

        dirindices = np.argsort(np.abs(p2))

        h = vecs[:,dirindices[2]]
        l = vecs[:,dirindices[1]]
        m = vecs[:,dirindices[0]]

# rotate the helix itself into the new coordinates 
# in many cases, this seems to not be a a great rotation.  It seems to
# start tilting the helix a bit in many cases. Not sure why!!!!

        S = np.zeros([3,3])
        S = vecs
        cp = np.dot(c,S)
        cp1 = cp[:,dirindices]
        cp = cp1

        up = np.dot(u,S)
        up1 = up[:,dirindices]
        up = up1

        up2 = np.diff(cp,axis=0)
        for upi in up2:
            upi /= np.sqrt(np.dot(upi,upi))

        FQ = np.zeros([3,3])
        for upi in up:
            FQ += 1.5*np.outer(upi,upi)
        FQ /= len(up)
        FQ -= 0.5*np.eye(3)

        avecos = np.mean(up[:,2])
        avesin = np.sqrt(1-avecos**2)

        zaxis = np.array([0,0,1])
        upr = np.zeros(np.shape(u))
        for i in range(np.shape(upr)[0]):
         scal = np.sqrt(1-up[i,2]**2)
         # project out into x,y plane
         ax1 = np.array([up[i,0]/scal,up[i,1]/scal,0])
         # normal from the plane 
         nm = np.cross(zaxis,ax1) # the normal to the plane
         v = up[i] # the vector to rotate
         # R(theta)v = nm(nm.v) + cos(theta) (nm x v) x nm + sin(-theta)(nm x v)  # from wikipedia
         upr[i]= nm*np.dot(nm,v) + avecos*np.cross(np.cross(nm,v),nm) - avesin*np.cross(nm,v)

        #cmid = 0.5*(cp[0:-1,:] + cp[1:,:])
        #z = cmid[0:,2]/length

        curves = [c,cp,u,up,upr]
        labels = ['helix (unrotated)', 'helix (rotated)', 'directors (unrotated)', 'directors (rotated)', 'directors (rotated to helix)']
        for i in range(len(curves)):
         fig = plt.figure(i)
         curve = curves[i]
         label = labels[i]
         ax = fig.gca(projection='3d')
         ax.plot(curve[:,0], curve[:,1], curve[:,2], label=label)
         ax.legend()
         plt.xlabel('x')
         plt.ylabel('y')
         #plt.zlabel('z') # not defined?
         plt.show()

        return(p2)


def get_helical_data(cgmodel):
        """
        """
        cgmodel = orient_along_z_axis(cgmodel) 

        # Get the new backbone particle positions
        backbone_positions = []
        for particle in range(len(cgmodel.positions)):
          if cgmodel.get_particle_type(particle) == "backbone":
            backbone_positions.append(cgmodel.positions[particle])
        backbone_positions = np.array([[float(i.in_units_of(unit.angstrom)._value) for i in coord] for coord in backbone_positions])

        # radius
        axis_distances = []
        rotations = 0.0
        for position in backbone_positions:
          axis_distance = distance(unit.Quantity([float(position[0]),0.0,0.0],unit.angstrom),unit.Quantity(position,unit.angstrom))
          axis_distances.append(axis_distance)
          if len(axis_distances) > 1:
            rotation = np.arctan(position[1]/position[2]) - last_angle
            last_angle = rotation
            rotations = rotations + rotation
          else:
            rotation = np.arctan(position[1]/position[2])
            last_angle = rotation
            rotations = rotations + rotation
            
        radius = mean(np.array([float(dist.in_units_of(unit.angstrom)._value) for dist in axis_distances]))
        particles_per_turn = float(cgmodel.polymer_length/(rotations/6.28))

        # pitch
        #
        # Shift all coordinates so that the first backbone atom has z=0

        shift = - cgmodel.positions[0][2]._value

        axis_deltas = []
        for position in cgmodel.positions:
          position[0]._value = position[0]._value + shift
          if abs(position[0]._value - cgmodel.positions[0][0]._value) > 0:
            axis_deltas.append(float(position[0]._value-cgmodel.positions[0][0]._value))
        average_delta = mean(axis_deltas)
        pitch = average_delta * particles_per_turn

        return(radius,pitch,particles_per_turn)
        




import numpy as np
from math import exp, log
# OpenMM utilities
import mdtraj as md
import simtk.unit as unit
import pymbar
from pymbar import timeseries
kB = unit.Quantity(0.008314462,unit.kilojoule_per_mole)  #Boltzmann constant (Gas constant) in kJ/(mol*K)

def bin_samples(sample_kn,n_bins):
        """
        """
        max_value,min_value=None,None
        for index_1 in range(len(sample_kn)):
          for index_2 in range(len(sample_kn[index_1])):
            sample = sample_kn[index_1][index_2]
            if max_value == None: max_value = sample
            if min_value == None: min_value = sample
            if sample > max_value: max_value = sample
            if sample < min_value: min_value = sample

        bin_size = (max_value-min_value)/(n_bins+1)
        bins = np.array([[min_value+i*bin_size,min_value+(i+1)*bin_size] for i in range(n_bins)])
        bin_counts = np.zeros((len(sample_kn),len(bins)))
        
        for index_1 in range(len(sample_kn)):
          for index_2 in range(len(sample_kn[index])):
            sample = sample_kn[index_1][index_2]

        return(bins,bin_counts)

def get_decorrelated_samples(replica_positions,replica_energies,temperature_list):
        """
        Given a set of replica exchange trajectories, energies, and associated temperatures, this function returns decorrelated samples, as obtained from pymbar with timeseries.subsampleCorrelatedData.

        :param replica_positions: Positions array for the replica exchange data for which we will write PDB files
        :type replica_positions: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( np.array( [n_replicas,cgmodel.num_beads,3] ), simtk.unit )

        :param replica_energies: List of dimension num_replicas X simulation_steps, which gives the energies for all replicas at all simulation steps 
        :type replica_energies: List( List( float * simtk.unit.energy for simulation_steps ) for num_replicas )

        :param temperature_list: List of temperatures for the simulation data.
        :type temperature_list: List( float * simtk.unit.temperature )

        :returns:
           - configurations ( List( `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ (n_decorrelated_samples,cgmodel.num_beads,3), simtk.unit ) ) - A list of decorrelated samples
           - energies ( List( `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ) ) - The energies for the decorrelated samples (configurations)

        """
        all_poses = []
        all_energies = []

        for replica_index in range(len(replica_positions)):
          energies = replica_energies[replica_index][replica_index]
          [t0,g,Neff_max] = timeseries.detectEquilibration(energies)
          energies_equil = energies[t0:]
          poses_equil = replica_positions[replica_index][t0:]
          indices = timeseries.subsampleCorrelatedData(energies_equil)
          for index in indices:
            all_energies.append(energies_equil[index])
            all_poses.append(poses_equil[index])

        all_energies = np.array([float(energy) for energy in all_energies])

        return(all_poses,all_energies)

def get_entropy_differences(mbar):
        """
        Given an `MBAR <https://pymbar.readthedocs.io/en/latest/mbar.html>`_ class object, this function computes the entropy differences for the states defined within.

        :param mbar: An MBAR() class object (from the 'pymbar' package)
        :type mbar: `MBAR <https://pymbar.readthedocs.io/en/latest/mbar.html>`_ class object

        :returns:
          - Delta_s ( np.array( n_mbar_states x n_thermo_states ) - Entropy differences for the thermodynamic states in 'mbar'
          - dDelta_s ( np.array( n_mbar_states x n_thermo_states ) - Uncertainty in the entropy differences for the thermodynamic states in 'mbar'

        """
        results = mbar.computeEntropyAndEnthalpy()
        results = {'Delta_f': results[0], 'dDelta_f': results[1], 'Delta_u': results[2], 'dDelta_u': results[3], 'Delta_s': results[4], 'dDelta_s': results[5]}
        Delta_s = results['Delta_s']
        dDelta_s = results['dDelta_s']
        return(Delta_s,dDelta_s)

def get_enthalpy_differences(mbar):
        """
        Given an `MBAR <https://pymbar.readthedocs.io/en/latest/mbar.html>`_ class object, this function computes the enthalpy differences for the states defined within.

        :param mbar: An MBAR() class object (from the 'pymbar' package)
        :type mbar: `MBAR <https://pymbar.readthedocs.io/en/latest/mbar.html>`_ class object

        :returns:
          - Delta_u ( np.array( n_mbar_states x n_thermo_states ) - Enthalpy differences for the thermodynamic states in 'mbar'
          - dDelta_u ( np.array( n_mbar_states x n_thermo_states ) - Uncertainty in the enthalpy differences for the thermodynamic states in 'mbar'

        """
        results = mbar.computeEntropyAndEnthalpy()
        results = {'Delta_f': results[0], 'dDelta_f': results[1], 'Delta_u': results[2], 'dDelta_u': results[3], 'Delta_s': results[4], 'dDelta_s': results[5]}
        Delta_u = results['Delta_u']
        dDelta_u = results['dDelta_u']
        return(Delta_u,dDelta_u)

def get_free_energy_differences(mbar):
        """
        Given an `MBAR <https://pymbar.readthedocs.io/en/latest/mbar.html>`_ class object, this function computes the free energy differences for the states defined within.

        :param mbar: An MBAR() class object (from the 'pymbar' package)
        :type mbar: `MBAR <https://pymbar.readthedocs.io/en/latest/mbar.html>`_ class object

        :returns:
          - df_ij ( np.array( n_mbar_states x n_thermo_states ) - Free energy differences for the thermodynamic states in 'mbar'
          - ddf_ij ( np.array( n_mbar_states x n_thermo_states ) - Uncertainty in the free energy differences for the thermodynamic states in 'mbar'

        """
        results = mbar.getFreeEnergyDifferences()
        results = {'Delta_f': results[0], 'dDelta_f': results[1]}
        df_ij = results['Delta_f']
        ddf_ij = results['dDelta_f']
        return(df_ij,ddf_ij)

def calc_temperature_spacing(min_temp,max_temp,num_replicas,replica_index):
        """
        Given the parameters to define a temperature range as input, as well as the current temperature index, this function uses logarithmic scaling to calculate the temperature spacing.

        :param min_temp: The minimum temperature in the temperature list.
        :type min_temp: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :param max_temp: The maximum temperature in the temperature list.
        :type max_temp: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :param num_replicas: The number of temperatures in the list.
        :type num_replicas: int

        :param replica_index: The index for the current temperature of interest.
        :type replica_index: int

        :returns:
          - delta ( `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ) - The spacing to use when assigning the next temperature value.
        """
        T_replica_index = min_temp * exp( replica_index * log( max_temp._value / min_temp._value )/(num_replicas-1) )
        T_previous_index = min_temp * exp( (replica_index-1) * log( max_temp._value / min_temp._value )/(num_replicas-1) )
        delta = T_replica_index - T_previous_index 
        return(delta)

def get_temperature_list(min_temp,max_temp,num_replicas):
        """
        Given the parameters to define a temperature range as input, this function uses logarithmic spacing to generate a list of intermediate temperatures.

        :param min_temp: The minimum temperature in the temperature list.
        :type min_temp: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :param max_temp: The maximum temperature in the temperature list.
        :type max_temp: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :param num_replicas: The number of temperatures in the list.
        :type num_replicas: int

        :returns:
           - temperature_list ( List( float * simtk.unit.temperature ) ) - List of temperatures

        """
        temperature_list = []
        temperature_list.append(min_temp)
        replica_index = 1
        while len(temperature_list) > 0 and temperature_list[-1].__lt__(max_temp):
          delta = calc_temperature_spacing(min_temp,max_temp,num_replicas,replica_index)
          last_temperature = temperature_list[-1]
          temperature = last_temperature.__add__(delta)
          temperature_list.append(temperature)
          replica_index = replica_index + 1
        return(temperature_list)

def get_intermediate_temperatures(T_from_file,NumIntermediates):
        """
        Given a list of temperatures and a number of intermediate states as input, this function calculates the values for temperatures intermediate between those in this list, as the mean between values in the list.

        :param T_from_file: List of temperatures
        :type T_from_file: List( float * simtk.unit.temperature )

        :param NumIntermediates: The number of states to insert between existing states in 'T_from_file'
        :type NumIntermediates: int

        :returns:
           - Temp_k ( List( float * simtk.unit.temperature ) ) - A new list of temperatures that includes the inserted intermediates.

        """

        deltas = []
        for i in range(1,len(T_from_file)):
         deltas.append((T_from_file[i]._value-T_from_file[i-1]._value)/(NumIntermediates+1))
         deltas.append((T_from_file[i]._value-T_from_file[i-1]._value)/(NumIntermediates+1))
        originalK = len(T_from_file)

        Temp_k = []
        val_k = []
        current_T = min([T_from_file[i]._value for i in range(len(T_from_file))])

        for delta in deltas:
           current_T = current_T + delta
           Temp_k.append(current_T)

        if len(Temp_k) != (len(T_from_file) + NumIntermediates* (len(T_from_file)-NumIntermediates-1)):
          print("Error: new temperatures are not being assigned correctly.")
          print("There were "+str(len(T_from_file))+" temperatures before inserting intermediates,")
          print(str(NumIntermediates)+" intermediate strucutures were requested,")
          print("and there were "+str(len(Temp_k))+" temperatures after inserting intermediates.")
          exit()

        Temp_k = np.array([temp for temp in Temp_k])
        return(Temp_k)

def get_mbar_expectation(E_kln,temperature_list,NumIntermediates,output=None,mbar=None):
        """
        Given a properly-formatted matrix of energies with associated temperatures this function reweights with MBAR (if 'mbar'=None), and can also compute the expectation value for any property of interest.

        .. warning:: This function accepts an input matrix thtat has either 'E_kln' or 'E_kn' format, but always provides an 'E_kn'-formatted matrix in return.

        :param E_kln: A matrix of energies or samples for a property that we would like to use to make predictions with MBAR.
        :type E_kln: List( List( float * simtk.unit.energy for simulation_steps ) for num_replicas ) OR List( List( List( float * simtk.unit.energy for simulation_steps ) for num_replicas ) for num_replicas )

        :param temperature_list: List of temperatures for the simulation data.
        :type temperature_list: List( float * simtk.unit.temperature )

        :param NumIntermediates: The number of states to insert between existing states in 'T_from_file'
        :type NumIntermediates: int

        :param output: The 'output' option to use when calling MBAR, default = 'differences'
        :type output: str

        :param mbar: An MBAR() class object (from the 'pymbar' package), default = None
        :type mbar: `MBAR <https://pymbar.readthedocs.io/en/latest/mbar.html>`_ class object

        :returns:
          - mbar ( `MBAR <https://pymbar.readthedocs.io/en/latest/mbar.html>`_ ) - An MBAR() class object (from the 'pymbar' package)          
         - E_kn ( List( List( float * simtk.unit.energy for num_samples ) for num_replicas ) ) - A matrix of energies or samples for a property that we would like to use to make predictions with MBAR.
         - result ( List( List( float for num_samples ) for num_replicas ) - The MBAR expectation value for the energies and/or other samples that were provided.
         - dresult ( List( List( float for num_samples ) for num_replicas ) - The MBAR expectation value for the energies and/or other samples that were provided.
         - Temp_k ( List( float * simtk.unit.temperature ) ) - A new list of temperatures that includes the inserted intermediates.



        """

        if mbar == None:
         NumTemps = len(temperature_list) # Last TEMP # + 1 (start counting at 1)

         kB = unit.Quantity(0.008314462,unit.kilojoule_per_mole)  #Boltzmann constant (Gas constant) in kJ/(mol*K)
         T_from_file = np.array([temperature._value for temperature in temperature_list])
         E_from_file = E_kln
         originalK = len(T_from_file)
         N_k = np.zeros(originalK,np.int32)

         g = np.zeros(originalK,np.float64)
         for k in range(originalK):  # subsample the energies
          g[k] = pymbar.timeseries.statisticalInefficiency(E_from_file[k][k])
          indices = np.array(pymbar.timeseries.subsampleCorrelatedData(E_from_file[k][k],g=g[k])) # indices of uncorrelated samples
          N_k[k] = len(indices)
          E_from_file[k,k,0:N_k[k]] = E_from_file[k,k,indices]

         if NumIntermediates > 0:
           Temp_k = get_intermediate_temperatures(temperature_list,NumIntermediates)
         else:
           Temp_k = np.array([temperature._value for temperature in temperature_list])

         # Update number of states
         K = len(Temp_k)
         # Loop, inserting E's into blank matrix (leaving blanks only where new Ts are inserted)

         Nall_k = np.zeros([K], np.int32) # Number of samples (n) for each state (k) = number of iterations/energies

         try:
          E_kn = np.zeros([K,len(E_from_file[0][0])], np.float64)
          for k in range(originalK-1):
              E_kn[k+k*NumIntermediates,0:N_k[k]] = E_from_file[k,k,0:N_k[k]]
              Nall_k[k+k*NumIntermediates] = N_k[k]

          E_kn[-1][0:N_k[-1]] = E_from_file[-1][-1][0:N_k[-1]]
          Nall_k[-1] = N_k[-1]


         except:
          E_kn = np.zeros([K,len(E_from_file[0])], np.float64)
          for k in range(originalK):
              E_kn[k+k*NumIntermediates,0:N_k[k]] = E_from_file[k,0:N_k[k]]
              Nall_k[k+k*NumIntermediates] = N_k[k]

         beta_k = 1 / (kB._value * Temp_k)

         allE_expect = np.zeros([K], np.float64)
         allE2_expect = np.zeros([K], np.float64)
         dE_expect = np.zeros([K],np.float64)
         u_kn = np.zeros([K,sum(Nall_k)], np.float64) # u_kln is reduced pot. ener. of segment n of temp k evaluated at temp l
         #index = 0
         for k in range(K):
           index = 0
           for l in range(K):
              u_kn[k,index:index+Nall_k[l]] = beta_k[k] * E_kn[l,0:Nall_k[l]]
              index = index + Nall_k[l]

         #------------------------------------------------------------------------
         # Initialize MBAR
         #------------------------------------------------------------------------

         print("Initializing MBAR:")
         print("--K = number of Temperatures with data = %d" % (originalK))
         print("--L = number of total Temperatures = %d" % (K))
         print("--N = number of Energies per Temperature = %d" % (np.max(Nall_k)))

         mbar = pymbar.MBAR(u_kn, Nall_k, verbose=False, relative_tolerance=1e-12,initial_f_k = None)

         E_kn = u_kn  # not a copy, we are going to write over it, but we don't need it any more.
         for k in range(K):
               E_kn[k,:]*=beta_k[k]**(-1)  # get the 'unreduced' potential -- we can't take differences of reduced potentials because the beta is different; math is much more confusing with derivatives of the reduced potentials.

        else:

          E_kn = E_kln          
          Temp_k = temperature_list

        if output != None:
               results = mbar.computeExpectations(E_kn,output='differences', state_dependent = True)
               results = {'mu': results[0], 'sigma': results[1]}
               result = results['mu']
               dresult = results['sigma']
        else:
               results = mbar.computeExpectations(E_kn, state_dependent = True)
               results = {'mu': results[0], 'sigma': results[1]}
               result = results['mu']
               dresult = results['sigma']

        return(mbar,E_kn,result,dresult,Temp_k)



import simtk.unit as unit
from msmbuilder.cluster.kcenters import KCenters
import mdtraj as md
from foldamers.utilities.iotools import write_pdbfile_without_topology

def concatenate_trajectories(pdb_file_list,combined_pdb_file="combined.pdb"):
        """
        Given a list of PDB files, this function reads their coordinates, concatenates them, and saves the combined coordinates to a new file (useful for clustering with MSMBuilder).

        :param pdb_file_list: A list of PDB files to read and concatenate
        :type pdb_file_list: List( str )

        :param combined_pdb_file: The name of file/path in which to save a combined version of the PDB files, default = "combined.pdb"
        :type combined_pdb_file: str

        :returns:
          - combined_pdb_file ( str ) - The name/path for a file within which the concatenated coordinates will be written.

        """
        traj_list = []
        for pdb_file in pdb_file_list:
          traj = md.load(pdb_file)
          traj_list.append(traj)
        return(combined_pdb_file)

def get_cluster_centroid_positions(pdb_file,cgmodel,n_clusters=None):
        """
        Given a PDB file and coarse grained model as input, this function performs K-means clustering on the poses in the PDB file, and returns a list of the coordinates for the "centroid" pose of each cluster.

        :param pdb_file: The path/name of a file from which to read trajectory data
        :type pdb_file: str

        :param cgmodel: A CGModel() class object
        :type cgmodel: class

        :param n_clusters: The number of "bins" within which to cluster the poses from the input trajectory.
        :type n_clusters: int

        :returns:
          - centroid_positions ( List ( np.array( float * unit.angstrom ( num_particles x 3 ) ) ) ) - A list of the poses corresponding to the centroids of all trajectory clusters.

        """
        centroid_positions = []
        traj = md.load(pdb_file)
        print(traj.n_frames)
        exit()
        if n_clusters == None:
          n_clusters = 50
        cluster_list = KCenters(n_clusters=n_clusters,metric='rmsd')
        cluster_list.fit(traj)
        centroid_list = cluster_list.cluster_centers_
        centroid_index = 1
        for centroid in centroid_list:
          positions = centroid.xyz[0] * unit.nanometer
          cgmodel.positions = positions
          file_name = str("centroid_"+str(centroid_index)+".xyz")
          write_pdbfile_without_topology(cgmodel,file_name)
          centroid_positions.append(positions)
        return(centroid_positions)

def align_structures(reference_traj,target_traj):
        """
        Given a reference trajectory, this function performs a structural alignment for a second input trajectory, with respect to the reference.

        :param reference_traj: The trajectory to use as a reference for alignment.
        :type reference_traj: `MDTraj() trajectory <http://mdtraj.org/1.6.2/api/generated/mdtraj.Trajectory.html>`_

        :param target_traj: The trajectory to align with the reference.
        :type target_traj: `MDTraj() trajectory <http://mdtraj.org/1.6.2/api/generated/mdtraj.Trajectory.html>`_

        :returns:
          - aligned_target_traj ( `MDTraj() trajectory <http://mdtraj.org/1.6.2/api/generated/mdtraj.Trajectory.html>`_ ) - The coordinates for the aligned trajectory.

        """

        aligned_target_traj = target_traj.superpose(reference_traj)

        return(aligned_target_traj)




import matplotlib.pyplot as pyplot
import numpy as np

def plot_distribution(x_data,y_data,file_name=None,plot_type=None,multiple=False,legend=None,legend_title=None,plot_title=None):
        """
        """
        index = 0
        figure = pyplot.figure(1)
        if plot_type == "Torsions":
           if file_name == None:
             file_name = "torsion_distribution.png"
           pyplot.title("Torsion Distribution")
           pyplot.xlabel("Torsion Angle (Degrees)")
        if plot_type =="Angles":
           if file_name == None:
             file_name = "bond_angle_distribution.png"
           pyplot.title("Bond Angle Distribution")
           pyplot.xlabel("Bond Angle (Degrees)")
        
        pyplot.ylabel("Counts")
        if multiple:
         x_data = np.array([[float(n) for n in x] for x in x_data])
         y_data = np.array([[float(n) for n in y] for y in y_data])
         for x,y in zip(x_data,y_data):
          line, = pyplot.plot(x,y)
          if legend:
            line.set_label(str(legend[index]))
          index = index + 1
         if legend != None:
          if legend_title != None:
           pyplot.legend(title=legend_title)
          else:
           pyplot.legend()
        else:
          x_data = np.array([float(x) for x in x_data])
          y_data = np.array([float(y) for y in y_data])
          if legend != None:
            line, = pyplot.plot(x_data,y_data)
            line.set_label(str(legend[index]))
          else:
            pyplot.plot(x_data,y_data)
        if plot_title != None:
          pyplot.title(plot_title)
        pyplot.savefig(file_name)
        pyplot.close()

        return
